"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SkewHeap = void 0;
const utils_1 = require("src/tree/utils");
const binaryHeap_1 = require("./binaryHeap");
const utils_2 = require("./utils");
/**
 *
 */
class SkewHeap {
    /**
     * Instantiate a heap.
     *
     * @param compareFn - The function to determine the order of elements.
     * @param elements - A set of elements to initialize the list with.
     */
    constructor(compareFn, elements) {
        this.compare = compareFn;
        this.length = 0;
        this.addAll(elements ?? []);
    }
    addAll(elements) {
        for (const element of elements) {
            this.push(element);
        }
        return this.length;
    }
    clear() {
        this.length = 0;
        this.root = undefined;
    }
    comparator() {
        return this.compare;
    }
    contains(element) {
        for (const node of utils_1.preOrderTraverse(this.root)) {
            if (element === node.value) {
                return true;
            }
        }
        return false;
    }
    delete(element) {
        if (this.root == null) {
            return false;
        }
        if (this.root.value === element) {
            this.pop();
            return true;
        }
        for (const par of utils_1.preOrderTraverse(this.root)) {
            const key = par.left && par.left.value === element
                ? 'left'
                : par.right && par.right.value === element
                    ? 'right'
                    : undefined;
            if (key != null) {
                const node = par[key];
                par[key] = utils_2.skewMerge(this.compare, [node.left, node.right]);
                --this.length;
                return true;
            }
        }
        return false;
    }
    *dump() {
        for (const node of utils_1.preOrderTraverse(this.root)) {
            yield node.value;
        }
    }
    merge(heap) {
        if (heap instanceof SkewHeap) {
            this.root = utils_2.skewMerge(this.compare, [this.root, utils_1.clone(heap.root)]);
            this.length += heap.size;
        }
        else if (heap instanceof binaryHeap_1.BinaryHeap) {
            this.root = utils_2.skewMerge(this.compare, [this.root, utils_1.toBinaryTree(heap['array'])]);
            this.length += heap.size;
        }
        else {
            this.addAll(heap.dump());
        }
        return this;
    }
    peek() {
        return this.root?.value;
    }
    pop() {
        if (this.root == null) {
            return undefined;
        }
        const value = this.root.value;
        this.root = utils_2.skewMerge(this.compare, [this.root.left, this.root.right]);
        --this.length;
        return value;
    }
    push(value) {
        this.root = utils_2.skewMerge(this.compare, [this.root, { value }]);
        return ++this.length;
    }
    pushPop(value) {
        this.push(value);
        return this.pop();
    }
    replace(value) {
        if (this.root == null) {
            this.root = { value };
            this.length = 1;
            return undefined;
        }
        const out = this.root.value;
        this.root = utils_2.skewMerge(this.compare, [this.root.left, this.root.right, { value }]);
        return out;
    }
    get size() {
        return this.length;
    }
    /**
     * Iterate through the heap in sorted order.
     *
     * **Note:** Unexpected behavior can occur if the collection is modified during iteration.
     *
     * @returns An iterator through the heap.
     */
    *[Symbol.iterator]() {
        if (this.root == null) {
            return;
        }
        const heap = new SkewHeap((a, b) => this.compare(a.value, b.value), [
            this.root,
        ]);
        do {
            const node = heap.pop();
            yield node.value;
            node.left && heap.push(node.left);
            node.right && heap.push(node.right);
        } while (heap.size > 0);
    }
    update(curElement, newElement) {
        if (this.root == null) {
            return false;
        }
        if (this.root.value === curElement) {
            this.root = utils_2.skewMerge(this.compare, [
                this.root.left,
                this.root.right,
                { value: newElement },
            ]);
            return true;
        }
        for (const par of utils_1.preOrderTraverse(this.root)) {
            const key = par.left && par.left.value === curElement
                ? 'left'
                : par.right && par.right.value === curElement
                    ? 'right'
                    : undefined;
            if (key != null) {
                const node = par[key];
                par[key] = undefined;
                this.root = utils_2.skewMerge(this.compare, [
                    this.root,
                    node.left,
                    node.right,
                    { value: newElement },
                ]);
                return true;
            }
        }
        return false;
    }
}
exports.SkewHeap = SkewHeap;
//# sourceMappingURL=skewHeap.js.map